#!/usr/bin/env python3

from PIL import Image, ImageFilter, ImageEnhance, ImageOps
import subprocess
import sys
import os
import pytesseract
import cv2
import numpy as np

import argparse

def data_to_sentences(input_):
    # get all words
    words = []
    for ind, elm in enumerate(input_["text"]):
        if elm != "":
            word = {}
            word["text"] = elm

            word["block_index"] = input_["block_num"][ind]
            word["xy"] = (input_["left"][ind], input_["top"][ind])
            words.append(word)

    # put words into sentences
    length_ = max(input_["block_num"]) + 1
    sentences_list_uncollapsed = [[] for i in range(length_)]
    # print(sentences_list_uncollapsed)
    for word in words:
        sentences_list_uncollapsed[word["block_index"]].append(word)

    sentences_list = [{} for i in range(length_)]
    for e in sentences_list:
        e["sentence"] = ""
    for index, sent_ in enumerate(sentences_list_uncollapsed):
        once_ = True
        for word_ in sent_:
            sentences_list[index]["sentence"] += " " + (word_["text"])
            if once_:
                once_ = False
                sentences_list[index]["xy"] = word_["xy"]
                sentences_list[index]["sentence"] = sentences_list[index]["sentence"][
                    1:
                ]

    sentences_list = [elm for elm in sentences_list if elm["sentence"] != ""]

    # if you need to understand whats going on, just look at the intermediate
    # objects. It is just reformatting data..
    return sentences_list


parser = argparse.ArgumentParser(
    prog="img2ttxxtt",
    description="Convert various images to text, and use OCR to replace text in said images with actual, real text",
    epilog="Description",
)

parser.add_argument("backend", choices=["jp2a", "chafa"])
parser.add_argument("filename")
parser.add_argument("-D", "--dim", type=type((int, int)))
parser.add_argument("-W", "--width", type=int)
parser.add_argument("-H", "--height", type=int)
parser.add_argument(
    "-C", "--colors", choices=["2", "8", "16", "256", "true"], default="16"
)
parser.add_argument("-P","--process", type=bool,default=False,help="Enable extra processing - mostly for black on white text sparsly in the picture (such as comics or mangas)")


def parse_command(args):
    command = []
    got = parser.parse_args(args)
    command.append(got.backend)
    if got.backend == "jp2a":
        if got.dim:
            command.append("--size=" + str(got.dim[0]) + "x" + str(got.dim[1]))
        if got.width:
            command.append("--width=" + str(got.width))
        if got.height:
            command.append("--height=" + str(got.height))
        match got.colors:
            case "true":
                got.colors = "24"
            case "16":
                got.colors = "4"
            case "256":
                got.colors = "8"
            case "8":
                print("8 color-mode not supported by jp2a")
                exit(1)
        command.append("--fill")
        command.append("--color-depth=" + str(got.colors))

    if got.backend == "chafa":
        if got.dim:
            command.append("-s")
            command.append(str(got.dim[0]) + "x" + str(got.dim[1]))
        if got.width:
            command.append("-s")
            command.append(str(got.width))
        if got.height:
            print("height param unsupported by chafa")
            exit(1)
        match got.colors:
            case "256":
                got.colors = "240"  # ssssh, no one will ever know !!!!!!
            case "true":
                got.colors = "full"
        command.append("-c")
        command.append(str(got.colors))

    command.append(got.filename)
    return command


if __name__ == "__main__":
    PROCESS = parser.parse_args(sys.argv[1:]).process 
    DEBUG = False

    # wraps the call
    PIPE = subprocess.PIPE  # alias

    # TODO process command right
    backend = None
    command = parse_command(sys.argv[1:])
    sub_env = os.environ.copy()

    # get lines and col cuz it is shell dependent
    linz = subprocess.run(["tput", "lines"], stdout=PIPE, stderr=PIPE)
    colz = subprocess.run(["tput", "cols"], stdout=PIPE, stderr=PIPE)

    # i swear the str to int makes sense
    sub_env["LINES"] = str(int(linz.stdout))
    sub_env["COLUMNS"] = str(int(colz.stdout))

    # call command, and get the res (placed bbox overimg to txt)
    res = subprocess.run(
        command, stdout=PIPE, stderr=PIPE, universal_newlines=True, env=sub_env
    )

    # get picture's text with pytesseract
    im = Image.open(sys.argv[2])
    pwidth, pheight = im.size
    # we extract all text using textbounding box. We apply several imag
    # processing to get the best bbox. Tesseract is good enough if there is
    # mostly text on pic
    sub_images = []
    if PROCESS:
        # try to denoise and get lines
        # https://stackoverflow.com/questions/4632174/what-processing-steps-should-i-use-to-clean-photos-of-line-drawings
        denoised = im.filter(ImageFilter.BLUR).filter(
            ImageFilter.GaussianBlur(radius=1)
        )

        grey = im.convert("L")
        # grey = ImageEnhance.Brightness(grey).enhance(2)
        # grey = ImageEnhance.Contrast(grey).enhance(4)
        impix = im.load()
        depix = denoised.load()
        greypix = grey.load()
        for y in range(pheight):
            for x in range(pwidth):
                greypix[x, y] = min(
                    255,
                    max(
                        255 + impix[x, y][0] - depix[x, y][0],
                        255 + impix[x, y][1] - depix[x, y][1],
                        255 + impix[x, y][2] - depix[x, y][2],
                    ),
                )
        grey = grey.filter(ImageFilter.MinFilter(15))
        grey = grey.filter(ImageFilter.GaussianBlur(radius=5))
        # grey = ImageEnhance.Sharpness(grey).enhance(2)
        # grey = grey.filter(ImageFilter.EDGE_ENHANCE)
        grey = ImageEnhance.Contrast(grey).enhance(2)
        grey = ImageOps.invert(grey)
        # grey.save("out.jpg")

        # https://stackoverflow.com/questions/14134892/convert-image-from-pil-to-opencv-format
        cvgrey = cv2.cvtColor(np.array(grey, np.uint8), cv2.COLOR_RGB2BGR)
        cvgrey = cv2.cvtColor(cvgrey, cv2.COLOR_BGR2GRAY)
        ret, thresh = cv2.threshold(cvgrey, 245, 255, 0)
        # cv2.imwrite("outbb.jpg", cvgrey)
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
        dilated = cv2.dilate(thresh, kernel, iterations=20)
        contours, _ = cv2.findContours(dilated, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

        im_bgred = cv2.cvtColor(np.array(im), cv2.COLOR_RGB2GRAY)
        for c in contours:
            rect = cv2.boundingRect(c)
            # Todo check for too small rect, delete
            x, y, w, h = rect
            cv2.rectangle(im_bgred, (x, y), (x + w, y + h), (255, 255, 0), 2)
            cropped = im_bgred[y : y + h, x : x + w]
            # print (cv2.contourArea(c))
            # print(len(cropped))
            # print(im_bgred)
            cropped_bright = ImageEnhance.Brightness(Image.fromarray(cropped)).enhance(
                2
            )
            #cropped_bright.save(str(x) + "outc.jpg")
            sub_images.append({"xy": (x, y), "img": cropped_bright})

        #cv2.imwrite("outbb.jpg", im_bgred)
        # im = grey

    # Now we pass every sub_images through tesseract :)
    # notes that I correct the xy
    all_sentences = []
    for im_ in sub_images:
        subpic_info = pytesseract.image_to_data(
            im_["img"], output_type=pytesseract.Output.DICT, config=r"-l eng --dpi 300"
        )
        # print(subpic_info)
        sentences_list = data_to_sentences(subpic_info)
        for sentence_ in sentences_list:
            sentence_["xy"] = (
                sentence_["xy"][0] + im_["xy"][0],
                sentence_["xy"][1] + im_["xy"][1],
            )
        all_sentences += sentences_list

    if not (PROCESS):
        all_sentences = data_to_sentences(
            pytesseract.image_to_data(
                im, output_type=pytesseract.Output.DICT, config=r"-l eng --dpi 300"
            )
        )

    if DEBUG:
        print(res.stderr)
        print(all_sentences)
        exit(1)

    if command[0] == "jp2a":  # TODO fix 2 color case
        cheight = res.stdout.count("\n")
        cwidth = int(res.stdout.count("\033[0m") / cheight)  # on how jp2a works

    if command[0] == "chafa":
        cheight = res.stdout.count("\n")
        cwidth = int((res.stdout.count("\033") - cheight) / cheight)

    # actually print the results and the texts
    print("\033[2J", end="")  # clears screen
    print(res.stdout, end="\033[0m")
    for sentence in all_sentences:
        cx = int((sentence["xy"][0] / pwidth) * cwidth)
        cy = int((sentence["xy"][1] / pheight) * cheight)
        print(f"\033[{cy};{cx}H{sentence['sentence']}")
    print("\033[u")
    # print(sentences_list)
